/*
 =========================================================================================================
 C. Voglis, Department of Computer Science,
 Univeristy of Ioannina, Greece

 File: 				SOURCE_local_opt.c
 Usage: 			Implementation of the local search driver
 Last Modified: 	21/08/2013
 =========================================================================================================
 */

#include <stdlib.h>
#include <string.h>
#include "SOURCE_params.h"
#include "SOURCE_local_opt.h"
#if defined(_OPENMP)
#include <omp.h>
#else
#define omp_get_thread_num()	0
#endif

extern void Objective_F(double X[], int N, double *F);
void parse_torczon_args(char *args, Torczon *torc);
void parse_bfgs_args(char *args, Bfgs *torc);

/*
 ------------------------------------------------------
 PERFORM LOCAL OPTIMIZATION BY CALLING APPROPRIATE LOCAL SEARCHES
 ------------------------------------------------------

 Input: 
 double	 	*in_x 	- Starting point
 int 		n	- Problem's dimension
 double		*xll	- Problem's lower bounds
 double		*xrl	- Problem's upper bounds


 Output:
 double 		*out_x	- Ending  point of the local minimization
 double 		*out_val- Objective function value at "out_x"
 double 		*grms	- Euclidean norm of the gradient vector at "out_x"
 int 			*fc	- Number of function evaluations performed
 int 			*gc	- Number of gradient evaluations performed
 */
void local_optimization(double *inout_x, int n, double *out_val, double *grms,
		double *xll, double *xrl, int *local, long *fc, long *gc) {

	int i, NF, NG, NOC;
	double VAL, *TP;
	double GRMS;
	double h__[n * n];
	int i__;
	double y[n], gn[n], gs[n], xs[n];
	int nod;
	double val, rho, grad[n];
	int marg[n], info, iter;
	double ftol;
	int ixat[n];
	double gtol, xtol, sigma;
	int nocex, nogex, itmax;
	double comacc;
	int igradu = 0;
	int i__1;
	int nomaxc, lsmeth;
	int ihessu = 0, lsiter, iprint, iord;
	int which, term;

	TP = (double *) malloc(n * sizeof(double));
	GRMS = 1.0;
	NG = 0;
	NF = 0;
	which = algo.method;

	for (i = 0; i < n; i++) {
		TP[i] = inout_x[i];
	}

	Objective_F(TP, n, &VAL);

	/*
	 *
	 Call local optimization subroutines
	 *
	 */
	if (which == 0) {
#ifdef DEBUG
		printf("-------------------------------------------------------\n");
		printf("Calling MDS with (%d, %d, %f, %f, %f) \n", algo.torc.maxfevals, algo.torc.maxiter, algo.torc.mu,
				algo.torc.theta, algo.torc.delta);
		printf("input point\n");
		for (i__ = 0; i__ < n; ++i__) {
			printf("%f ", TP[i__]);
		}
		printf("\n");
		printf("Val %f \n", VAL);
#endif
		mds(TP, n, &VAL, EPS, algo.torc.maxfevals, algo.torc.maxiter, algo.torc.mu,
				algo.torc.theta, algo.torc.delta, &NF, xll, xrl, &term);
#ifdef DEBUG
		printf("-------------------------------------------------------\n");
		printf("Called  MDS \n");
		printf("output point\n");
		for (i__ = 0; i__ < n; ++i__) {
			printf("%f ", TP[i__]);
		}
		printf("\n");
		printf("Val %f \n", VAL);
#endif
		if (term == 2)
			GRMS = 0.0;
	} else {
		memset(h__, 0, n * n * sizeof(double));
		nod = n;
		comacc = 1e-15f;

		itmax = algo.bfgs.maxiter;
		nomaxc = algo.bfgs.maxfevals;
		gtol = algo.bfgs.geps;
		xtol = algo.bfgs.xeps;
		ftol = algo.bfgs.feps;
		lsiter = algo.bfgs.lsiter;
		rho = algo.bfgs.rho;
		sigma = algo.bfgs.sigma;
		iord = algo.bfgs.order;

		lsmeth = 0;
		iprint = 1;
		i__1 = nod;
		for (i__ = 1; i__ <= i__1; ++i__) {
			ixat[i__ - 1] = 1;
			marg[i__ - 1] = 0;
		}
		val = VAL;

		int anal = 1;
		if (algo.method == 1)
			anal = 0;
		else if (algo.method == 2)
			anal = 1;
#ifdef DEBUG
		printf("-------------------------------------------------------\n");
		printf("Calling BFGS with anal = %d and order = %d\n", anal, iord);
		printf("input point\n");
		for (i__ = 0; i__ < nod; ++i__) {
			printf("%f ", TP[i__]);
		}
		printf("\n");
		printf("Val %f \n", val);
#endif
		bfgsmv_(h__, TP, grad, &nod, &val, &nomaxc, &comacc, &gtol, &itmax, &xtol, &lsmeth, &rho,
				&sigma, &lsiter, &ftol, &info, &iter, &nocex, &nogex, &iprint, ixat, marg, xll,
				xrl, &igradu, &ihessu, xs, gn, y, gs, anal, iord);
#ifdef DEBUG
		printf("Called BFGS with anal = %d\n", anal);
		printf("output point\n");
		for (i__ = 0; i__ < nod; ++i__) {
			printf("%f ", TP[i__]);
		}
		printf("\n");
		printf("Val %f \n", val);
		printf("Stats: Info = %d, Fevals=%d, Gevals=%d \n", info, nocex, nogex);
		printf("-------------------------------------------------------\n");

#endif

		VAL = val;
		NF = nocex;
		NG = nogex;
		if (info != 7 && info != 3)
			GRMS = 0.0;

	}

	NOC = 0;

	for (i = 0; i < n; i++) {
		inout_x[i] = TP[i];
	}

#ifdef TINKER
	int NOD = n;
	int step;
	for (i=0; i<NOD; i++)
	{
		if (inout_x[i] > 1e6)
		{
			step = floor(inout_x[i] / 360.0);
			inout_x[i] = inout_x[i] - (step-1)*360.0;
		}
		while (inout_x[i] > 180.0)
		inout_x[i] = inout_x[i] - 360.0;
	}
	for (i=0; i<NOD; i++)
	{
		if (inout_x[i] < -1e6)
		{
			step = floor(fabs(inout_x[i]) / 360.0);
			inout_x[i] = inout_x[i] + (step-1)*360.0;
		}
		while (inout_x[i] < -180.0)
		inout_x[i] = inout_x[i] + 360.0;
	}
#endif
	*out_val = VAL;
	*grms = GRMS;
#pragma omp critical
	{
		*fc = *fc + NF + NOC;
		*gc = *gc + NG;
	}

	if (GRMS < GEPS)
		*local = 1;
	else
		*local = 0;

	free(TP);
}

void parse_bfgs_args(char *args, Bfgs *torc) {
	char *token;
	char left[32];
	double right;

	token = strtok(args, " ");
	while (token != NULL) {
		if (sscanf(token, "%[A-Za-z]=%lf", left, &right) != EOF) {
			if (strcmp(left, "feps") == 0)
				torc->feps = (right);
			if (strcmp(left, "xeps") == 0)
				torc->xeps = (right);
			if (strcmp(left, "geps") == 0)
				torc->geps = (right);
			if (strcmp(left, "rho") == 0)
				torc->rho = (right);
			if (strcmp(left, "sigma") == 0)
				torc->sigma = (right);
			if (strcmp(left, "maxiter") == 0)
				torc->maxiter = floor(right);
			if (strcmp(left, "maxfevals") == 0)
				torc->maxfevals = floor(right);
			if (strcmp(left, "lsiter") == 0)
				torc->lsiter = floor(right);
			if (strcmp(left, "order") == 0) {
				torc->order = floor(right);
				if (torc->order != 1 && torc->order != 2)
					torc->order = 1;
				printf("order = %i\n", torc->order);
			}
		}
		token = strtok(NULL, " ");
	}
}
void parse_torczon_args(char *args, Torczon *torc) {
	char *token;
	char left[32];
	double right;
	torc->mu = 2.0;
	torc->theta = 0.5;
	torc->delta = 0.5;

	token = strtok(args, " ");
	while (token != NULL) {
		if (sscanf(token, "%[A-Za-z]=%lf", left, &right) != EOF) {
			if (strcmp(left, "mu") == 0)
				torc->mu = (right);
			if (strcmp(left, "theta") == 0)
				torc->theta = (right);
			if (strcmp(left, "delta") == 0)
				torc->delta = (right);
			if (strcmp(left, "maxiter") == 0)
				torc->maxiter = floor(right);
			if (strcmp(left, "maxfevals") == 0)
				torc->maxfevals = floor(right);
		}
		token = strtok(NULL, " ");
	}

}

