/*
 =========================================================================================================
 C. Voglis, Department of Computer Science,
 Univeristy of Ioannina, Greece

 File: 				lj.c
 Usage: 			Implementation of Lennard-Jones interatomic potential. (includes derivative)
 Last Modified: 	21/08/2013
 =========================================================================================================
 */
#include <math.h>
#include "objective.h"

const double pi = 3.14159265358979;


double finmin(double[], int);
double ljpot(double);
double gljpot(double);
void glj(double p[], int n, double gp[]);


double lj(double p[], int n) {
	int np, i, j, i3, j3, np1;
	double v, dx, dy, dz, rij;
	double xi, yi, zi;
	double xj, yj, zj;

	np = n / 3;

	v = 0.0;
	np1 = np - 1;
	for (i = 0; i < np1; ++i) {
		i3 = 3 * i;
		xi = p[i3];
		yi = p[i3 + 1];
		zi = p[i3 + 2];
		for (j = i + 1; j < np; ++j) {
			j3 = 3 * j;
			xj = p[j3];
			yj = p[j3 + 1];
			zj = p[j3 + 2];
			dx = xi - xj;
			dy = yi - yj;
			dz = zi - zj;
			rij = sqrt(dx * dx + dy * dy + dz * dz);
			v += ljpot(rij);
		}
	}

	return v;
}

void glj(double p[], int n, double gp[]) {
	int np, i, j, i3, j3, np1;
	double v, dx, dy, dz, rij;
	double xi, yi, zi;
	double xj, yj, zj;

	for (i = 0; i < n; i++)
		gp[i] = 0.0;

	np = n / 3;

	v = 0.0;
	np1 = np - 1;
	for (i = 0; i < np1; ++i) {
		i3 = 3 * i;
		xi = p[i3];
		yi = p[i3 + 1];
		zi = p[i3 + 2];
		for (j = i + 1; j < np; ++j) {
			j3 = 3 * j;
			xj = p[j3];
			yj = p[j3 + 1];
			zj = p[j3 + 2];
			dx = xi - xj;
			dy = yi - yj;
			dz = zi - zj;
			rij = sqrt(dx * dx + dy * dy + dz * dz);
			v = gljpot(rij);
			gp[i3] += 2.0 * dx * v / rij;
			gp[i3 + 1] += 2.0 * dy * v / rij;
			gp[i3 + 2] += 2.0 * dz * v / rij;
			gp[j3] -= 2.0 * dx * v / rij;
			gp[j3 + 1] -= 2.0 * dy * v / rij;
			gp[j3 + 2] -= 2.0 * dz * v / rij;

		}
	}
	for (i = 0; i < n; i++)
		gp[i] /= 2.0;
}

#define SIGMA    1.0
#define EPSILON4 4.0
double ljpot(double r) {
	double sr, sr3, sr6, sr12;

	sr = SIGMA / r;
	sr3 = sr * sr * sr;
	sr6 = sr3 * sr3;
	sr12 = sr6 * sr6;
	return EPSILON4 * (sr12 - sr6);
}

double gljpot(double r) {
	double sr, sr3, sr6, sr12;

	sr = SIGMA / r;
	sr3 = sr * sr * sr;
	sr6 = (sr3 * sr3);
	sr12 = (sr6 * sr6);
	return EPSILON4 * (6.0 * sr6 / r - 12.0 * sr12 / r);
}

void Objective_F(double X[], int N, double *F) {
	*F = lj(X, N);
	FEVALCOUNT

}

void Bounds_F(double *xl, double *xr, int N) {
	int i;
	for (i = 0; i < N; i++) {
		xl[i] = -2.0;
		xr[i] = 2.0;
	}
}

void Objective_G(double X[], int N, double G[]) {
	glj(X, N, G);
}

